// Copyright 6/4/2019 Varjo Technologies Oy. All Rights Reserved.

#include "VarjoEyeTrackerModule.h"
#include "GameFramework/HUD.h"
#include "VarjoGaze.h"
#include "Engine/Engine.h"
#include "VarjoEyeTrackerTypes.h"

IMPLEMENT_MODULE(FVarjoEyeTrackerModule, VarjoEyeTracker)

static TAutoConsoleVariable<int32> CVarEnableEyetrackingDebug(TEXT("Varjo.debug.EnableEyetrackingDebug"), 1, TEXT("0 - Eyetracking debug visualizations are disabled. 1 - Eyetracking debug visualizations are enabled."));

/************************************************************************/
/* FVarjoEyeTrackerModule                                                     */
/************************************************************************/
void FVarjoEyeTrackerModule::StartupModule()
{
	IEyeTrackerModule::StartupModule();
	VarjoEyeTracker = TSharedPtr<FVarjoEyeTracker, ESPMode::ThreadSafe>(new FVarjoEyeTracker());
	OnDrawDebugHandle = AHUD::OnShowDebugInfo.AddRaw(this, &FVarjoEyeTrackerModule::OnDrawDebug);
}

void FVarjoEyeTrackerModule::ShutdownModule()
{
	AHUD::OnShowDebugInfo.Remove(OnDrawDebugHandle);
	FCoreUObjectDelegates::PreLoadMap.Remove(OnPreLoadMapHandle);
}

TSharedPtr<class IEyeTracker, ESPMode::ThreadSafe> FVarjoEyeTrackerModule::CreateEyeTracker()
{
	VarjoEyeTracker->Reset();
	return VarjoEyeTracker;
}

void FVarjoEyeTrackerModule::OnDrawDebug(AHUD* HUD, UCanvas* Canvas, const FDebugDisplayInfo& DisplayInfo, float& YL, float& YPos)
{
	VarjoEyeTracker->m_VREyeTracker->EnableDebugView(CVarEnableEyetrackingDebug.GetValueOnGameThread());
}

bool FVarjoEyeTrackerModule::IsEyeTrackerConnected() const
{
	IEyeTracker* EyeTracker = VarjoEyeTracker.Get();
	if (EyeTracker)
	{
		EEyeTrackerStatus Status = EyeTracker->GetEyeTrackerStatus();
		if (/*(Status != EEyeTrackerStatus::NotTracking) && */(Status != EEyeTrackerStatus::NotConnected))
		{
			return true;
		}
	}

	return false;
}
