// Copyright 6/4/2019 Varjo Technologies Oy. All Rights Reserved.

#pragma once

#include "Kismet/BlueprintFunctionLibrary.h"
#include "EyeTrackerFunctionLibrary.h"
#include "IMotionController.h"
#include "VarjoEyeTrackerTypes.h"
#include "VarjoEyeTrackerFunctionLibrary.generated.h"

UCLASS(ClassGroup = Varjo)
class VARJOEYETRACKER_API UVarjoEyeTrackerFunctionLibrary : public UEyeTrackerFunctionLibrary
{
	GENERATED_UCLASS_BODY()

public:
	UFUNCTION(BlueprintPure, BlueprintPure, Category = "EyeTracking|Varjo")
	static EVarjoEyeTrackingCalibrationStatus GetCalibrationStatus();

	/**
	* Requests to start the gaze calibration
	*/
	UFUNCTION(BlueprintCallable, Category = "EyeTracking|Varjo")
	static bool RequestGazeCalibration();

	/**
	* Returns the ray pointing from the current EyeTracker position to the gaze direction
	*
	* @param     start			(out) The point where the ray starts
	* @param     direction		(out) The ray direction
	* @param	 focusDistance	(out) The estimated gaze direction focus point distance
	*/
	UFUNCTION(BlueprintPure, Category = "EyeTracking|Varjo")
	static void GetGazeRay(FVector& start, FVector& direction, float& focusDistance);

	/**
	* Returns the ray pointing from the current EyeTracker position to the gaze direction
	*
	* @param     return varjo gata data list
	*/
	UFUNCTION(BlueprintPure, Category = "EyeTracking|Varjo")
	static TArray<FString> GetGazeStringArray();

	/**
	* Returns the gaze data list
	*
	* @param     return varjo gata data list
	*/
	UFUNCTION(BlueprintPure, Category = "EyeTracking|Varjo")
	static TArray<FVarjoEyeTrackingData> GetGazeArray();

	UFUNCTION(BlueprintCallable, Category = "EyeTracking|Varjo")
	static bool SaveGazeData(FString SaveDirectory, FString FileName, const TArray<FVarjoEyeTrackingData>& gazeArray);


	/**
	* Returns the gaze data
	*
	* @param     return varjo gaze data
	*/
	UFUNCTION(BlueprintPure, Category = "EyeTracking|Varjo")
	static bool GetEyeTrackerGazeData(FVarjoEyeTrackingData& eyeTrackingData);
};
