// Copyright 6/4/2019 Varjo Technologies Oy. All Rights Reserved.

#pragma once

#include "HeadMountedDisplayBase.h"
#include "XRRenderBridge.h"

#include "Windows/AllowWindowsPlatformTypes.h"

// Varjo API
#include "Varjo.h"
#include "Varjo_d3d11.h"

// D3D11 Specific
#include <d3dcommon.h>
#include <d3dcompiler.h>
#pragma comment(lib, "d3dcompiler.lib")
#pragma comment(lib, "d3d11.lib")

// Other
#include <windows.h>

#include "Windows/HideWindowsPlatformTypes.h"

#include "VarjoTextureSet.h"

class VarjoCustomPresent : public FXRRenderBridge
{
public:
	VarjoCustomPresent(class FVarjoHMD* varjoHMD);
	virtual ~VarjoCustomPresent();

	virtual void varjoInit() = 0;
	virtual void varjoSubmit() = 0;
	bool isInitialized() const;
	void OnBackBufferResize() override;
	bool Present(int& SyncInterval) override;
	virtual void BeginRendering();
	virtual void FinishRendering();
	virtual void UpdateViewport(const FViewport& Viewport, FRHIViewport* InViewportRHI) = 0;
	virtual void SetNeedReinitRendererAPI();
	virtual bool NeedsNativePresent() override;
	virtual void Reset();
	virtual void Shutdown();
	void PostPresent();
	virtual bool CreateRenderTargetTexture(FTexture2DRHIRef& OutTargetableTexture, FTexture2DRHIRef& OutShaderResourceTexture) = 0;
	virtual void AliasTextureResources(FTextureRHIParamRef DestTexture, FTextureRHIParamRef SrcTexture) = 0;
	void setResolutionFraction(float resolutionFraction) { m_resolutionFraction = resolutionFraction; };
	float getResolutionFraction() const { return m_resolutionFraction; };
	bool getButtonEvent(int& button, bool& pressed);
	void renderOcclusionMesh(FRHICommandList& RHICmdList, int viewIndex);
	void handleVarjoEvents(UGameViewportClient* gameViewportClient);
	void getFocusViewPosAndSize(EStereoscopicPass stereoPass, float& x, float& y, float& width, float& height) const;

protected:
	virtual FTextureRHIRef CreateTexture(ID3D11Texture2D* d3dTexture) const = 0;
	VarjoTextureSetPtr CreateTextureSet();
	ID3D11Device* m_device;
	ID3D11DeviceContext* m_deviceContext;
	varjo_Session* m_session;
	ID3D11Texture2D* m_texture;
	VarjoTextureSetPtr m_textureSet;
	float m_resolutionFraction = 1.0f;
	bool m_inFrame = false;

	// Varjo API related
	varjo_GraphicsInfo* m_graphicsInfo;
	varjo_FrameInfo* m_frameInfo;
	varjo_SubmitInfo* m_submitInfo;

private:
	void setupOcclusionMeshes();

	varjo_Event* m_event;
	TArray<FTextureRHIRef> CreateTextures(const TArray<ID3D11Texture2D*>& d3dTextures) const;
	class FVarjoHMD* m_varjoHMD;
	varjo_Mesh2Df* m_varjoOcclusionMesh;
	FHMDViewMesh m_occlusionMeshes[4];
	bool m_buttonEventExists = false;
	varjo_EventButton m_buttonEvent;

	// Normalized position and size of focus views, with respect to context views.
	float m_focusX[2]{ 0.0f };
	float m_focusY[2]{ 0.0f };
	float m_focusWidth[2]{ 1.0f };
	float m_focusHeight[2]{ 1.0f };
};
