// Copyright 6/4/2019 Varjo Technologies Oy. All Rights Reserved.

#pragma once

#include "Kismet/BlueprintFunctionLibrary.h"
#include "VarjoVRControllerLibrary.generated.h"

/** Defines the set of input events you want associated with your VarjoVR d-pad buttons */
UENUM(BlueprintType)
enum class EVarjoVRTouchDPadMapping : uint8
{
	/** Maps each controller's four touchpad buttons to the respective FaceButton1/2/3/4 events */
	FaceButtons,

	/** Maps each controller's four touchpad buttons to the respective Thumbstick_Up/Down/Left/Right events  */
	ThumbstickDirections,

	/** Unmaps all directional touchpad input so that the d-pad buttons won't trigger their own input events */
	Disabled
};

/** */
UCLASS()
class VARJOVRCONTROLLER_API UVarjoVRControllerLibrary : public UBlueprintFunctionLibrary
{
	GENERATED_BODY()

public:

	/**
	 * Sets which input events you want associated with the VarjoVR controller's 
	 * directional touchpad buttons - does so for both left and right controllers.
	 *
	 * @param  NewMapping	Defines the set of input events you want associated with the four directional touchpad buttons.
	 */
	UFUNCTION(BlueprintCallable, Category="VarjoHMD")
	static void SetTouchDPadMapping(EVarjoVRTouchDPadMapping NewMapping = EVarjoVRTouchDPadMapping::FaceButtons); // @see VarjoVRController.cpp for implementation

	/**
	* Given a controller index and a hand, returns a boolean which tells if the controller is connected to the computer
	*
	* @param    ControllerIndex Index of the controller to get the tracked device ID for
	* @param    MotionSource	Which hand's controller to get the position and orientation for
	* @return   True if the specified controller is connected
	*/
	UFUNCTION(BlueprintPure, Category = "MotionController|VarjoHMD")
	static bool IsControllerConnected(int32 ControllerIndex, FName MotionSource);

	/**
	* Given a controller index and a hand, returns the battery level
	*
	* @param    ControllerIndex Index of the controller to get the tracked device ID for
	* @param    MotionSource	Which hand's controller to get the position and orientation for
	* @return   The battery level of the controller
	*/
	UFUNCTION(BlueprintPure, Category = "MotionController|VarjoHMD")
	static float GetControllerBatteryLevel(int32 ControllerIndex, FName MotionSource);
};
